"use strict";
/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
Object.defineProperty(exports, "__esModule", { value: true });
const assert_1 = require("assert");
const testClientLogger_1 = require("./testClientLogger.cjs");
const ClientIds = ["A", "B", "C", "D"];
class ReconnectTestHelper {
    constructor() {
        this.clients = (0, testClientLogger_1.createClientsAtInitialState)({ initialState: "" }, ...ClientIds);
        this.logger = new testClientLogger_1.TestClientLogger(this.clients.all);
        this.ops = [];
        this.perClientOps = this.clients.all.map(() => []);
        this.seq = 0;
    }
    idxFromName(name) {
        return name.charCodeAt(0) - "A".charCodeAt(0);
    }
    insertText(clientName, pos, text) {
        const client = this.clients[clientName];
        this.ops.push(client.makeOpMessage(client.insertTextLocal(pos, text), ++this.seq));
    }
    removeRange(clientName, start, end) {
        const client = this.clients[clientName];
        this.ops.push(client.makeOpMessage(client.removeRangeLocal(start, end), ++this.seq));
    }
    obliterateRange(clientName, start, end) {
        const client = this.clients[clientName];
        this.ops.push(client.makeOpMessage(client.obliterateRangeLocal(start, end), ++this.seq));
    }
    disconnect(clientNames) {
        const clientIdxs = clientNames.map(this.idxFromName);
        this.ops
            .splice(0)
            .forEach((op) => this.clients.all.forEach((c, i) => clientIdxs.includes(i) ? this.perClientOps[i].push(op) : c.applyMsg(op)));
    }
    processAllOps() {
        this.ops.splice(0).forEach((op) => this.clients.all.forEach((c) => c.applyMsg(op)));
    }
    reconnect(clientNames) {
        const clientIdxs = clientNames.map(this.idxFromName);
        this.perClientOps.forEach((clientOps, i) => {
            if (clientIdxs.includes(i)) {
                clientOps.splice(0).forEach((op) => this.clients.all[i].applyMsg(op));
            }
        });
    }
    submitDisconnectedOp(clientName, op) {
        const client = this.clients[clientName];
        const pendingSegmentGroups = client.peekPendingSegmentGroups();
        (0, assert_1.strict)(pendingSegmentGroups);
        this.ops.push(client.makeOpMessage(client.regeneratePendingOp(op, pendingSegmentGroups), ++this.seq));
    }
}
describe("obliterate", () => {
    it("obliterate does not expand during rebase", () => {
        const helper = new ReconnectTestHelper();
        helper.insertText("B", 0, "ABCD");
        helper.processAllOps();
        helper.removeRange("B", 0, 3);
        helper.disconnect(["C"]);
        const cOp = helper.clients.C.obliterateRangeLocal(0, 1);
        (0, assert_1.strict)(cOp);
        helper.reconnect(["C"]);
        helper.submitDisconnectedOp("C", cOp);
        helper.processAllOps();
        assert_1.strict.equal(helper.clients.A.getText(), "D");
        helper.logger.validate();
    });
    it.skip("deletes reconnected insert into obliterate range", () => {
        const helper = new ReconnectTestHelper();
        helper.insertText("B", 0, "ABCD");
        helper.processAllOps();
        helper.obliterateRange("B", 0, 3);
        helper.disconnect(["C"]);
        const cOp = helper.clients.C.insertTextLocal(2, "aaa");
        (0, assert_1.strict)(cOp);
        helper.reconnect(["C"]);
        helper.submitDisconnectedOp("C", cOp);
        helper.processAllOps();
        assert_1.strict.equal(helper.clients.A.getText(), "D");
        helper.logger.validate();
    });
    it("does not delete reconnected insert at start of obliterate range", () => {
        const helper = new ReconnectTestHelper();
        helper.insertText("B", 0, "ABCD");
        helper.processAllOps();
        helper.obliterateRange("B", 0, 3);
        helper.disconnect(["C"]);
        const cOp = helper.clients.C.insertTextLocal(0, "aaa");
        (0, assert_1.strict)(cOp);
        helper.reconnect(["C"]);
        helper.submitDisconnectedOp("C", cOp);
        helper.processAllOps();
        assert_1.strict.equal(helper.clients.A.getText(), "aaaD");
        helper.logger.validate();
    });
    it("does not delete reconnected insert at end of obliterate range", () => {
        const helper = new ReconnectTestHelper();
        helper.insertText("B", 0, "ABCD");
        helper.processAllOps();
        helper.obliterateRange("B", 0, 3);
        helper.disconnect(["C"]);
        const cOp = helper.clients.C.insertTextLocal(3, "aaa");
        (0, assert_1.strict)(cOp);
        helper.reconnect(["C"]);
        helper.submitDisconnectedOp("C", cOp);
        helper.processAllOps();
        assert_1.strict.equal(helper.clients.A.getText(), "aaaD");
        helper.logger.validate();
    });
});
//# sourceMappingURL=obliterate.reconnect.spec.cjs.map