"use strict";
/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.checkTextMatchRelative = exports.TestServer = void 0;
const collections_1 = require("../collections/index.cjs");
const mergeTreeNodes_1 = require("../mergeTreeNodes.cjs");
const mergeTree_1 = require("../mergeTree.cjs");
const textSegment_1 = require("../textSegment.cjs");
const MergeTreeTextHelper_1 = require("../MergeTreeTextHelper.cjs");
const testClient_1 = require("./testClient.cjs");
/**
 * Server for tests.  Simulates client communication by directing placing
 * messages in client queues.
 */
class TestServer extends testClient_1.TestClient {
    constructor(options) {
        super(options);
        this.seq = 1;
        this.clients = [];
        this.messageListeners = []; // Listeners do not generate edits
        this.clientSeqNumbers = new collections_1.Heap([], mergeTree_1.clientSeqComparer);
        this.upstreamMap = new collections_1.RedBlackTree(mergeTreeNodes_1.compareNumbers);
        this.minSeq = 0;
    }
    addUpstreamClients(upstreamClients) {
        // Assumes addClients already called
        this.upstreamMap = new collections_1.RedBlackTree(mergeTreeNodes_1.compareNumbers);
        for (const upstreamClient of upstreamClients) {
            this.clientSeqNumbers.add({
                refSeq: upstreamClient.getCurrentSeq(),
                clientId: upstreamClient.longClientId ?? "",
            });
        }
    }
    addClients(clients) {
        this.clientSeqNumbers = new collections_1.Heap([], mergeTree_1.clientSeqComparer);
        this.clients = clients;
        for (const client of clients) {
            this.clientSeqNumbers.add({
                refSeq: client.getCurrentSeq(),
                clientId: client.longClientId ?? "",
            });
        }
    }
    addListeners(listeners) {
        this.messageListeners = listeners;
    }
    applyMsg(msg) {
        super.applyMsg(msg);
        if (testClient_1.TestClient.useCheckQ) {
            // eslint-disable-next-line @typescript-eslint/no-unnecessary-type-assertion
            const clid = this.getShortClientId(msg.clientId);
            return checkTextMatchRelative(msg.referenceSequenceNumber, clid, this, msg);
        }
        else {
            return false;
        }
    }
    // TODO: remove mappings when no longer needed using min seq
    // in upstream message
    transformUpstreamMessage(msg) {
        if (msg.referenceSequenceNumber > 0) {
            msg.referenceSequenceNumber =
                this.upstreamMap.get(msg.referenceSequenceNumber)?.data ?? 0;
        }
        msg.origin = {
            id: "A",
            sequenceNumber: msg.sequenceNumber,
            minimumSequenceNumber: msg.minimumSequenceNumber,
        };
        this.upstreamMap.put(msg.sequenceNumber, this.seq);
        msg.sequenceNumber = -1;
    }
    copyMsg(msg) {
        // eslint-disable-next-line @typescript-eslint/consistent-type-assertions
        return {
            clientId: msg.clientId,
            clientSequenceNumber: msg.clientSequenceNumber,
            contents: msg.contents,
            minimumSequenceNumber: msg.minimumSequenceNumber,
            referenceSequenceNumber: msg.referenceSequenceNumber,
            sequenceNumber: msg.sequenceNumber,
            type: msg.type,
        };
    }
    applyMessages(msgCount) {
        let _msgCount = msgCount;
        while (_msgCount > 0) {
            const msg = this.dequeueMsg();
            if (msg) {
                if (msg.sequenceNumber >= 0) {
                    this.transformUpstreamMessage(msg);
                }
                msg.sequenceNumber = this.seq++;
                msg.minimumSequenceNumber = this.minSeq;
                if (this.applyMsg(msg)) {
                    return true;
                }
                if (this.clients) {
                    let minCli = this.clientSeqNumbers.peek();
                    if (minCli &&
                        minCli.clientId === msg.clientId &&
                        minCli.refSeq < msg.referenceSequenceNumber) {
                        const cliSeq = this.clientSeqNumbers.get();
                        const oldSeq = cliSeq.refSeq;
                        cliSeq.refSeq = msg.referenceSequenceNumber;
                        this.clientSeqNumbers.add(cliSeq);
                        minCli = this.clientSeqNumbers.peek();
                        if (minCli.refSeq > oldSeq) {
                            msg.minimumSequenceNumber = minCli.refSeq;
                            this.minSeq = minCli.refSeq;
                        }
                    }
                    for (const client of this.clients) {
                        client.enqueueMsg(msg);
                    }
                    if (this.messageListeners) {
                        for (const listener of this.messageListeners) {
                            listener.enqueueMsg(this.copyMsg(msg));
                        }
                    }
                }
            }
            else {
                break;
            }
            _msgCount--;
        }
        return false;
    }
    incrementalGetText(start, end) {
        const range = { start, end };
        if (range.start === undefined) {
            range.start = 0;
        }
        if (range.end === undefined) {
            range.end = this.getLength();
        }
        const context = new textSegment_1.TextSegment("");
        const stack = new collections_1.Stack();
        const initialState = new mergeTreeNodes_1.IncrementalMapState(this.mergeTree.root, { leaf: incrementalGatherText }, 0, this.getCurrentSeq(), this.getClientId(), context, range.start, range.end, 0);
        stack.push(initialState);
        while (!stack.empty()) {
            this.mergeTree.incrementalBlockMap(stack);
        }
        return context.text;
    }
}
exports.TestServer = TestServer;
function incrementalGatherText(segment, state) {
    if (textSegment_1.TextSegment.is(segment)) {
        if (state.start <= 0 && state.end >= segment.text.length) {
            state.context.text += segment.text;
        }
        else {
            state.context.text +=
                state.end >= segment.text.length
                    ? segment.text.substring(state.start)
                    : segment.text.substring(state.start, state.end);
        }
    }
    state.op = mergeTreeNodes_1.IncrementalExecOp.Go;
}
/**
 * Used for in-memory testing.  This will queue a reference string for each client message.
 */
function checkTextMatchRelative(refSeq, clientId, server, msg) {
    const client = server.clients[clientId];
    const serverText = new MergeTreeTextHelper_1.MergeTreeTextHelper(server.mergeTree).getText(refSeq, clientId);
    const cliText = client.checkQ.shift()?.data;
    if (cliText === undefined || cliText !== serverText) {
        console.log(`mismatch `);
        console.log(msg);
        //        console.log(serverText);
        //        console.log(cliText);
        // eslint-disable-next-line @typescript-eslint/no-base-to-string
        console.log(server.mergeTree.toString());
        // eslint-disable-next-line @typescript-eslint/no-base-to-string
        console.log(client.mergeTree.toString());
        return true;
    }
    return false;
}
exports.checkTextMatchRelative = checkTextMatchRelative;
//# sourceMappingURL=testServer.cjs.map